package adaptivehac;

import java.util.Arrays;
import java.util.Iterator;

public class Cluster implements Iterable<Element> 
{
  static public Cluster EMPTY_CLUSTER_;
  static public Pair<Cluster> EMPTY_PAIR_;
  static {
    EMPTY_CLUSTER_ = new Cluster();
    EMPTY_PAIR_ = Pair.create(EMPTY_CLUSTER_, EMPTY_CLUSTER_);
  }

  // The elements contained in this cluster.
  final protected Element[] elements_;
  
  // The agglomerative clustering history.
  final public Cluster c1_;
  final public Cluster c2_;
  
  // Clustering index.
  public int index_;
  
  public Cluster()
  {
    elements_ = new Element[0];
    c1_ = c2_ = null;
    index_ = -1;
  }
  
  public Cluster(Cluster c)
  {
    elements_ = c.elements_;
    c1_ = c.c1_;
    c2_ = c.c2_;
    index_ = c.index_;
  }
  
  public Cluster(Element el)
  {
    elements_ = new Element[1];
    elements_[0] = el;
    c1_ = c2_ = null;
    index_ = -1;
  }
  
  public Cluster(Element[] elements, Cluster c1, Cluster c2)
  {
    if (elements != null)
      Arrays.sort(elements);
    
    elements_ = elements;
    c1_ = c1;
    c2_ = c2;
    index_ = -1;
  }
  
  public Object getFirstTrueID()
  {
    if (elements_.length == 0)
      return null;
    
    return elements_[0].getClusterID();
  }
  
  public int size()
  {
    return elements_.length;
  }
  
  public Element get(int i)
  {
    return elements_[i];
  }
  
  public String toString()
  {
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < elements_.length; i++)
      sb.append(elements_[i].toString()).append(" ");
    
    return sb.toString();
  }
  
  public boolean equals(Object o)
  {
    Cluster ec = (Cluster) o;
    
    return elements_ == ec.elements_;
  }

  
  public Iterator<Element> iterator()
  {
    return Arrays.asList(elements_).iterator();
  }
  
  

}
