package adaptivehac;

import java.util.Iterator;
import java.util.Vector;

public class Clustering implements Iterable<Cluster>
{
  protected Vector<Cluster> clusters_;
  transient TriangularMatrix<FeatureVector> fvs_;
  
  public Clustering()
  {
    clusters_ = new Vector<Cluster>();
    fvs_ = null;
  }
  
  public void add(Cluster cluster)
  {
    cluster.index_ = clusters_.size();
    clusters_.add(cluster);
  }

  public int size()
  {
    return clusters_.size();
  }
  
  public Cluster get(int i)
  {
    return clusters_.get(i);
  }
 
  public Iterator<Cluster> iterator()
  {
    return clusters_.iterator();
  }
  
  
  public Cluster join(Cluster c1, Cluster c2)
  {
    Cluster cl1 = c1.index_ < c2.index_ ? c1 : c2;
    Cluster cl2 = c1.index_ < c2.index_ ? c2 : c1;
    int i1 = cl1.index_;
    int i2 = cl2.index_;
    
    Element[] elements = new Element[cl1.size() + cl2.size()];
    for (int i = 0; i < cl1.size(); i++)
      elements[i] = cl1.elements_[i];
    for (int i = 0; i < cl2.size(); i++)
      elements[cl1.size() + i] = cl2.elements_[i];
    
    // Create new cluster from the two clusters.
    Cluster cluster = new Cluster(elements, null, null);
    cluster.index_ = cl1.index_;
    
    // Replace cl1 with new cluster.
    clusters_.set(cl1.index_, cluster);
   
    // Remove cluster cl2.
    clusters_.remove(cl2.index_);

    // Update cluster indeces.
    for (int i = cl2.index_; i < clusters_.size(); i++)
      clusters_.get(i).index_--;
    
    // Shift feature vectors in the right positions.
    if (fvs_ != null) {
      TriangularMatrix<FeatureVector> fvs = new TriangularMatrix<FeatureVector>(clusters_.size());
      for (int i = 0; i < i2; i++) {
        for (int j = i + 1; j < i2; j++)
          fvs.set(i, j, fvs_.get(i, j));
        for (int j = i2; j < clusters_.size(); j++)
          fvs.set(i, j, fvs_.get(i, j + 1));
      }
      for (int i = i2; i < clusters_.size(); i++) {
        for (int j = i + 1; j < clusters_.size(); j++)
          fvs.set(i, j, fvs_.get(i + 1, j + 1));
      }
      // Set feature vectors for pairs with new cluster to null.
      for (int i = 0; i < i1; i++)
        fvs.set(i, i1, null);
      for (int i = i1 + 1; i < clusters_.size(); i++)
        fvs.set(i1, i, null);
      fvs_ = fvs;
    }
      
    return cluster;
  }

  public void setFeatureVectors(Document doc, AlphabetHAC alphabet)
  {
    fvs_ = new TriangularMatrix<FeatureVector>(clusters_.size());
    
    for (int i = 0; i < clusters_.size(); i++) {
      for (int j = i + 1; j < clusters_.size(); j++) {
        Pair<Cluster> pair = Pair.create(clusters_.get(i), clusters_.get(j));
        fvs_.set(i, j, doc.getFeatureVector(pair, alphabet));
      }
    }
  }

  public void updateFeatureVectors(Document doc, AlphabetHAC alphabet, Cluster c)
  {
    for (int i = 0; i < c.index_; i++) {
      Pair<Cluster> pair = new Pair<Cluster>(clusters_.get(i), c);
      fvs_.set(i, c.index_, doc.getFeatureVector(pair, alphabet));
    }
    for (int j = c.index_ + 1; j < clusters_.size(); j++) {
      Pair<Cluster> pair = new Pair<Cluster>(c, clusters_.get(j));
      fvs_.set(c.index_, j, doc.getFeatureVector(pair, alphabet));
    }
  }
  
  public void setFeatureVectors(Document doc, AlphabetHAC alphabet, RankingModel rm)
  {
    fvs_ = new TriangularMatrix<FeatureVector>(clusters_.size());
    
    for (int i = 0; i < clusters_.size(); i++) {
      for (int j = i + 1; j < clusters_.size(); j++) {
        Pair<Cluster> pair = Pair.create(clusters_.get(i), clusters_.get(j));
        fvs_.set(i, j, doc.getFeatureVector(pair, alphabet));
      }
    }
  }

  public void updateFeatureVectors(Document doc, AlphabetHAC alphabet, Cluster c, RankingModel rm)
  {
    for (int i = 0; i < c.index_; i++) {
      Pair<Cluster> pair = new Pair<Cluster>(clusters_.get(i), c);
      fvs_.set(i, c.index_, doc.getFeatureVector(pair, alphabet));
    }
    for (int j = c.index_ + 1; j < clusters_.size(); j++) {
      Pair<Cluster> pair = new Pair<Cluster>(c, clusters_.get(j));
      fvs_.set(c.index_, j, doc.getFeatureVector(pair, alphabet));
    }
  }
  
}
