package adaptivehac;

import java.io.Serializable;
import java.util.Vector;

public class LinearDiscriminant implements Discriminant, Serializable
{
  private static final long serialVersionUID = 1;
  /*
  private static final int SIZE_ = 10000;

  public double[] w_;
  
  public LinearDiscriminant()
  {
    w_ = new double[SIZE_];
    Arrays.fill(w_, 0);
  }
  
  public LinearDiscriminant(LinearDiscriminant ld)
  {
    w_ = new double[ld.w_.length];
    for (int i = 0; i < w_.length; i++)
      w_[i] = ld.w_[i];
  }
  
  public void add(FeatureVector fv, double rate)
  {
    for (Feature f : fv.features_) {
      if (f.index_ == 0 || f.index_ >= w_.length) {
        System.out.println("Too many features!");
        System.exit(1);
      }
      w_[f.index_] = w_[f.index_] + rate * f.value_;
    }
  }
  
  public void sub(FeatureVector fv, double rate)
  {
    for (Feature f : fv.features_) {
      if (f.index_ == 0 || f.index_ >= w_.length) {
        System.out.println("Too many features!");
        System.exit(1);
      }
      w_[f.index_] = w_[f.index_] - rate * f.value_;
    }
  }
  
  public void scale(double factor)
  {
    for (int i = 0; i < w_.length; i++)
      w_[i] = w_[i] * factor;
  }
  
  public void add(LinearDiscriminant disc)
  {
    if (w_.length < disc.w_.length) {
      System.out.println("Too many features!");
      System.exit(1);
    }
    
    for (int i = 0; i < disc.w_.length; i++)
      w_[i] = w_[i] + disc.w_[i];
  }
  
  
  @Override
  public double evaluate(FeatureVector fv)
  {
    double result = 0;
    for (Feature f : fv.features_) {
      if (f.index_ == 0 || f.index_ >= w_.length) {
        System.out.println("Too many features!");
        System.exit(1);
      }
      result += w_[f.index_] * f.value_;
    }

    return result;
  }
*/
  public Vector<Double> w_;
  
  public LinearDiscriminant()
  {
    w_ = new Vector<Double>();
  }
  
  public LinearDiscriminant(LinearDiscriminant ld)
  {
    w_ = new Vector<Double>();
    w_.addAll(ld.w_);
  }
  
  public void add(FeatureVector fv, double rate)
  {
    for (Feature f : fv.features_) {
      if (w_.size() <= f.index_)
        growWeights(f.index_ + 1);
      w_.set(f.index_, w_.get(f.index_) + rate * f.value_);
    }
  }
  
  public void sub(FeatureVector fv, double rate)
  {
    for (Feature f : fv.features_) {
      if (w_.size() <= f.index_)
        growWeights(f.index_ + 1);
      w_.set(f.index_, w_.get(f.index_) - rate * f.value_);
    }
  }
  
  public void scale(double factor)
  {
    for (int i = 0; i < w_.size(); i++)
      w_.set(i, w_.get(i) * factor);
  }
  
  protected void growWeights(int size)
  {
    while (w_.size() < size)
      w_.add(0.0);
  }
  
  public void add(LinearDiscriminant disc)
  {
    if (w_.size() < disc.w_.size())
      growWeights(disc.w_.size());
    
    for (int i = 0; i < disc.w_.size(); i++)
      w_.set(i, w_.get(i) + disc.w_.get(i));
  }
  
  
  @Override
  public double evaluate(FeatureVector fv)
  {
    double result = 0;
    for (Feature f : fv.features_) {
      if (w_.size() <= f.index_)
        growWeights(f.index_ + 1);
      result += w_.get(f.index_) * f.value_;
    }

    return result;
  }

}
