package adaptivehac;

import java.util.Arrays;
import java.util.Vector;


public class OnlineMaxHAC extends AdaptiveHAC
{
  private static final long serialVersionUID = 1L;
  static public int NUM_EPOCHS_ = 10;
  static public int MAX_DISTANCE_ = Integer.MAX_VALUE;//50;
  
  public OnlineMaxHAC(PerceptronRankingModel rm)
  {
    rm_ = rm;
  }
  
  public OnlineMaxHAC(MiraRankingModel rm)
  {
    rm_ = rm;
  }
  
  public Pair<? extends Cluster> getEmptyPair()
  {
    return ElementCluster.EMPTY_PAIR_;
  }
  
  public void train(Document[] docs)
  {
    PerceptronRankingModel rm = (PerceptronRankingModel) rm_;
    // MiraRankingModel rm = (MiraRankingModel) rm_;
    
    for (int epoch = 0; epoch < NUM_EPOCHS_; epoch++) {
      System.out.println("Training epoch " + epoch + ".");
      for (Document doc : docs) {
        System.out.println("\tDocument " + doc.id_);
        updateModel(doc);
      }
    }

    rm.setAverageDiscriminant();
    rm.alphabet_.stopGrowth();
    rm.printFeatureWeights();
  }
  
  
  protected void updateModel(Document doc)
  {
    PerceptronRankingModel rm = (PerceptronRankingModel) rm_;
    // MiraRankingModel rm = (MiraRankingModel) rm_;

    boolean[] solved = new boolean[doc.elements_.length];
    Arrays.fill(solved, false);
    Clustering clustering = doc.getSingletonClusters();
    
    int decisions = 0;
    Pair<ElementCluster> best_pos = getBestPositivePair(doc, clustering, solved);
    Pair<ElementCluster> best_pair = getBestPair(doc, clustering, solved);
    while (best_pos != null) {
      // Best pair should be positive.
      if (!best_pair.equals(best_pos)) {
        rm.update(doc, Pair.create(best_pos, best_pair));
        // rm.updateAvg();
        // return;
      }
      
      rm.updateAvg();
      
      int anaphor = Math.max(best_pos.getLeft().getElement().index_, best_pos.getRight().getElement().index_);
      solved[anaphor] = true;
      clustering.join(best_pos.getLeft(), best_pos.getRight());
      
      if (++decisions % 10 == 0)
        System.out.println("\t  " + decisions + " clustering decisions.");
      
      best_pos = getBestPositivePair(doc, clustering, solved);
      best_pair = getBestPair(doc, clustering, solved);
    }
    
    // Best pair should be empty.
    if (!best_pair.equals(ElementCluster.EMPTY_PAIR_))
      rm.update(doc, Pair.create(ElementCluster.EMPTY_PAIR_, best_pair));
    rm.updateAvg();
  }
  
  
  private Pair<ElementCluster> getBestPositivePair(Document doc, Clustering clustering, boolean[] solved)
  {
    Pair<ElementCluster> best_pair = null;
    double best_score = Double.NEGATIVE_INFINITY;
    for (int i = 0; i < clustering.clusters_.size(); i++) {
      Cluster c1 = clustering.clusters_.get(i);
      for (int j = i + 1; j < clustering.clusters_.size(); j++) {
        Cluster c2 = clustering.clusters_.get(j);

        // Temporary distance constraint for efficiency.
        int distance = getElemDistance(c1, c2);
        if (distance > MAX_DISTANCE_)
          continue;
        
        if (c1.getFirstTrueID() == c2.getFirstTrueID()) {
          Pair<ElementCluster> c_pair =
            new Pair<ElementCluster>(new ElementCluster(c1, -1),
                                     new ElementCluster(c2, -1));
          double c_score = rm_.evaluate(doc, c_pair);
          for (int l1 = 0; l1 < c1.elements_.length; l1++) {
            for (int l2 = 0; l2 < c2.elements_.length; l2++) {
              int anaphor = Math.max(c1.elements_[l1].index_, c2.elements_[l2].index_);
              if (solved[anaphor])
                continue;
              Pair<ElementCluster> el_pair =
                new Pair<ElementCluster>(new ElementCluster(c1, l1),
                                         new ElementCluster(c2, l2));
              el_pair.getLeft().setIgnoreCluster(true);
              el_pair.getRight().setIgnoreCluster(true);
              double el_score = rm_.evaluate(doc, el_pair);
              double cel_score = c_score + el_score;
              if (cel_score > best_score) {
                Pair<ElementCluster> cel_pair =
                  new Pair<ElementCluster>(new ElementCluster(c1, l1),
                                           new ElementCluster(c2, l2));
                best_pair = cel_pair;
                best_score = cel_score;
              }
            }
          }
        }
      }
    }
    
    return best_pair;
  }


  private Pair<ElementCluster> getBestPair(Document doc, Clustering clustering, boolean[] solved)
  {
    Pair<ElementCluster> best_pair = ElementCluster.EMPTY_PAIR_;
    double best_score = rm_.evaluate(doc, ElementCluster.EMPTY_PAIR_);
    for (int i = 0; i < clustering.clusters_.size(); i++) {
      Cluster c1 = clustering.clusters_.get(i);
      for (int j = i + 1; j < clustering.clusters_.size(); j++) {
        Cluster c2 = clustering.clusters_.get(j);
        
        // Temporary distance constraint for efficiency.
        int distance = getElemDistance(c1, c2);
        if (distance > MAX_DISTANCE_)
          continue;
        
        Pair<ElementCluster> c_pair =
          new Pair<ElementCluster>(new ElementCluster(c1, -1),
                                   new ElementCluster(c2, -1));
        double c_score = rm_.evaluate(doc, c_pair);
        for (int l1 = 0; l1 < c1.elements_.length; l1++) {
          for (int l2 = 0; l2 < c2.elements_.length; l2++) {
            int anaphor = Math.max(c1.elements_[l1].index_, c2.elements_[l2].index_);
            if (solved[anaphor])
              continue;
            Pair<ElementCluster> el_pair =
              new Pair<ElementCluster>(new ElementCluster(c1, l1),
                                       new ElementCluster(c2, l2));
            el_pair.getLeft().setIgnoreCluster(true);
            el_pair.getRight().setIgnoreCluster(true);
            double el_score = rm_.evaluate(doc, el_pair);
            double cel_score = c_score + el_score;
            if (cel_score > best_score) {
              Pair<ElementCluster> cel_pair =
                new Pair<ElementCluster>(new ElementCluster(c1, l1),
                                         new ElementCluster(c2, l2));
              best_pair = cel_pair;
              best_score = cel_score;
            }
          }
        }
      }
    }
    
    return best_pair;
  }

  
  public Pair<ElementCluster> getBestNonEmptyPair(Document doc, Clustering clustering, boolean[] solved)
  {
    Pair<ElementCluster> best_pair = null;
    double best_score = Double.NEGATIVE_INFINITY;
    for (int i = 0; i < clustering.clusters_.size(); i++) {
      Cluster c1 = clustering.clusters_.get(i);
      for (int j = i + 1; j < clustering.clusters_.size(); j++) {
        Cluster c2 = clustering.clusters_.get(j);
        
        // Temporary distance constraint for efficiency.
        int distance = getElemDistance(c1, c2);
        if (distance > MAX_DISTANCE_)
          continue;
        
        Pair<ElementCluster> c_pair =
          new Pair<ElementCluster>(new ElementCluster(c1, -1),
                                   new ElementCluster(c2, -1));
        double c_score = rm_.evaluate(doc, c_pair);
        for (int l1 = 0; l1 < c1.elements_.length; l1++) {
          for (int l2 = 0; l2 < c2.elements_.length; l2++) {
            int anaphor = Math.max(c1.elements_[l1].index_, c2.elements_[l2].index_);
            if (solved[anaphor])
              continue;
            Pair<ElementCluster> el_pair =
              new Pair<ElementCluster>(new ElementCluster(c1, l1),
                                       new ElementCluster(c2, l2));
            el_pair.getLeft().setIgnoreCluster(true);
            el_pair.getRight().setIgnoreCluster(true);
            double el_score = rm_.evaluate(doc, el_pair);
            double cel_score = c_score + el_score;
            if (cel_score > best_score) {
              Pair<ElementCluster> cel_pair =
                new Pair<ElementCluster>(new ElementCluster(c1, l1),
                                         new ElementCluster(c2, l2));
              best_pair = cel_pair;
              best_score = cel_score;
            }
          }
        }
      }
    }
    
    return best_pair;
  }

  public int getElemDistance(Cluster c1, Cluster c2)
  {
    // Assert elements are sorted inside each cluster.
    int first1 = c1.elements_[0].index_;
    int last1 = c1.elements_[c1.elements_.length - 1].index_;
    
    int first2 = c2.elements_[0].index_;
    int last2 = c2.elements_[c2.elements_.length - 1].index_;
    
    if (first1 < first2) {
      if (last1 < first2)
        return first2 - last1;
      return 0;
    }
    else {
      if (last2 < first1)
        return first1 - last2;
      return 0;
    }
  }
  
  public Clustering cluster(Document doc)
  {
    boolean[] solved = new boolean[doc.elements_.length];
    Arrays.fill(solved, false);
    Clustering clustering = doc.getSingletonClusters();

    Pair<ElementCluster> best_pair;
    do {
      best_pair = getBestPair(doc, clustering, solved);
      if (best_pair != ElementCluster.EMPTY_PAIR_) {
        int anaphor = Math.max(best_pair.getLeft().getElement().index_, best_pair.getRight().getElement().index_);
        solved[anaphor] = true;
        clustering.join(best_pair.getLeft(), best_pair.getRight());
      }
    } 
    while (best_pair != ElementCluster.EMPTY_PAIR_);
        
    return clustering;
  }
  
  
  @SuppressWarnings("unused")
  private Pair<ElementCluster> getBestPairAll(Document doc, Clustering clustering, boolean[] solved)
  {
    Pair<ElementCluster> best_pair = ElementCluster.EMPTY_PAIR_;
    double best_score = rm_.evaluate(doc, ElementCluster.EMPTY_PAIR_);
    for (int i = 0; i < clustering.clusters_.size(); i++) {
      Cluster c1 = clustering.clusters_.get(i);
      for (int j = i + 1; j < clustering.clusters_.size(); j++) {
        Cluster c2 = clustering.clusters_.get(j);
        Vector<Pair<ElementCluster>> pairs = new Vector<Pair<ElementCluster>>();
        for (int l1 = 0; l1 < c1.elements_.length; l1++) {
          for (int l2 = l1 + 1; l2 < c1.elements_.length; l2++)
            pairs.add(Pair.create(new ElementCluster(c1, l1), 
                                  new ElementCluster(c1, l2)));
        }
        for (int l1 = 0; l1 < c2.elements_.length; l1++) {
          for (int l2 = l1 + 1; l2 < c2.elements_.length; l2++)
            pairs.add(Pair.create(new ElementCluster(c2, l1), 
                                  new ElementCluster(c2, l2)));
        }
        for (int l1 = 0; l1 < c1.elements_.length; l1++) {
          for (int l2 = 0; l2 < c2.elements_.length; l2++)
            pairs.add(Pair.create(new ElementCluster(c1, l1), 
                                  new ElementCluster(c2, l2)));
        }
    
        for (Pair<ElementCluster> pair : pairs) {
          ElementCluster ec1 = pair.getLeft();
          int l1 = ec1.element_;
          ElementCluster ec2 = pair.getRight();
          int l2 = ec2.element_;
          int anaphor = Math.max(ec1.elements_[l1].index_, ec2.elements_[l2].index_);
          if (solved[anaphor])
            continue;
          pair.getLeft().setIgnoreCluster(true);
          pair.getRight().setIgnoreCluster(true);
          double cel_score = rm_.evaluate(doc, pair);
          if (cel_score > best_score) {
            best_pair = pair;
            best_score = cel_score;
          }
        }
      }
    }
    
    return best_pair;
  }

  
  @SuppressWarnings("unused")
  private Pair<ElementCluster> getBestPositivePairAll(Document doc, Clustering clustering, boolean[] solved)
  {
    Pair<ElementCluster> best_pair = null;
    double best_score = Double.NEGATIVE_INFINITY;
    for (int i = 0; i < clustering.clusters_.size(); i++) {
      Cluster c1 = clustering.clusters_.get(i);
      for (int j = i + 1; j < clustering.clusters_.size(); j++) {
        Cluster c2 = clustering.clusters_.get(j);
        if (c1.getFirstTrueID() != c2.getFirstTrueID())
          continue;
        Vector<Pair<ElementCluster>> pairs = new Vector<Pair<ElementCluster>>();
        for (int l1 = 0; l1 < c1.elements_.length; l1++) {
          for (int l2 = l1 + 1; l2 < c1.elements_.length; l2++)
            pairs.add(Pair.create(new ElementCluster(c1, l1), 
                                  new ElementCluster(c1, l2)));
        }
        for (int l1 = 0; l1 < c2.elements_.length; l1++) {
          for (int l2 = l1 + 1; l2 < c2.elements_.length; l2++)
            pairs.add(Pair.create(new ElementCluster(c2, l1), 
                                  new ElementCluster(c2, l2)));
        }
        for (int l1 = 0; l1 < c1.elements_.length; l1++) {
          for (int l2 = 0; l2 < c2.elements_.length; l2++)
            pairs.add(Pair.create(new ElementCluster(c1, l1), 
                                  new ElementCluster(c2, l2)));
        }
    
        for (Pair<ElementCluster> pair : pairs) {
          ElementCluster ec1 = pair.getLeft();
          int l1 = ec1.element_;
          ElementCluster ec2 = pair.getRight();
          int l2 = ec2.element_;
          int anaphor = Math.max(ec1.elements_[l1].index_, ec2.elements_[l2].index_);
          if (solved[anaphor])
            continue;
          pair.getLeft().setIgnoreCluster(true);
          pair.getRight().setIgnoreCluster(true);
          double cel_score = rm_.evaluate(doc, pair);
          if (cel_score > best_score) {
            best_pair = pair;
            best_score = cel_score;
          }
        }
      }
    }
    
    return best_pair;
  }

  
  
}
