package adaptivehac;

import java.util.Random;
import java.util.Vector;

public class OnlineSumHAC extends AdaptiveHAC {

  private static final long serialVersionUID = 1L;
  static public int NUM_EPOCHS_ = 10;
  static public int MAX_DISTANCE_ = Integer.MAX_VALUE;//50;
  static public double EPSILON_ = 1E-10;
  
  private Random random_ = new Random(0);
  
  public OnlineSumHAC(PerceptronRankingModel rm)
  {
    rm_ = rm;
  }
  
  public OnlineSumHAC(MiraRankingModel rm)
  {
    rm_ = rm;
  }
  
  public void trainMira(Document[] docs)
  {
    MiraRankingModel rm = (MiraRankingModel) rm_;
    
    for (int epoch = 0; epoch < NUM_EPOCHS_; epoch++) {
      System.out.println("Training epoch " + epoch + ".");
      for (Document doc : docs) {
        System.out.println("\tDocument " + doc.id_);
        updateModelMira(doc);
      }
    }

    rm.setAverageDiscriminant();
    rm.alphabet_.stopGrowth();
    rm.printFeatureWeights();
  }
  
  
  protected void updateModelMira(Document doc)
  {
    MiraRankingModel rm = (MiraRankingModel) rm_;
    
    Clustering clustering = doc.getSingletonClusters();
    clustering.setFeatureVectors(doc, rm.alphabet_, rm);
    
    int decisions = 0;
    Pair<Cluster> best_pos = getBestPositivePair(doc, clustering);
    Pair<Cluster> best_pair = getBestPair(doc, clustering);
    while (best_pos != null) {
      // Best pair should be positive.
      if (!best_pair.equals(best_pos)) {
        rm.update(doc, Pair.create(best_pos, best_pair));
        // rm.updateAvg();
        // return;
      }
      
      rm.updateAvg();
      
      Cluster c = clustering.join(best_pos.getLeft(), best_pos.getRight());
      clustering.updateFeatureVectors(doc, rm.alphabet_, c, rm);
      
      if (++decisions % 10 == 0)
        System.out.println("\t  " + decisions + " clustering decisions.");
      
      best_pos = getBestPositivePair(doc, clustering);
      best_pair = getBestPair(doc, clustering);
    }
    
    // Best pair should be empty.
    if (best_pair != Cluster.EMPTY_PAIR_)
      rm.update(doc, Pair.create(Cluster.EMPTY_PAIR_, best_pair));
    rm.updateAvg();
  }
  
  
  public void train(Document[] docs)
  {
    PerceptronRankingModel rm = (PerceptronRankingModel) rm_;
    
    for (int epoch = 0; epoch < NUM_EPOCHS_; epoch++) {
      System.out.println("Training epoch " + epoch + ".");
      for (Document doc : docs) {
        System.out.println("\tDocument " + doc.id_);
        updateModelPerceptron(doc);
      }
    }

    rm.printFeatureWeights();

    rm.setAverageDiscriminant();
    rm.alphabet_.stopGrowth();
    rm.printFeatureWeights();
  }
  
  
  protected void updateModelPerceptron(Document doc)
  {
    PerceptronRankingModel rm = (PerceptronRankingModel) rm_;
    
    Clustering clustering = doc.getSingletonClusters();
    clustering.setFeatureVectors(doc, rm.alphabet_, rm);
    
    int decisions = 0;
    Pair<Cluster> best_pos = getBestPositivePair(doc, clustering);
    Pair<Cluster> best_pair = getBestPair(doc, clustering);
    while (best_pos != null) {
      // Best pair should be positive.
      if (!best_pair.equals(best_pos)) {
        rm.update(doc, Pair.create(best_pos, best_pair));
        // rm.updateAvg();
        // return;
      }
      
      rm.updateAvg();
      
      Cluster c = clustering.join(best_pos.getLeft(), best_pos.getRight());
      clustering.updateFeatureVectors(doc, rm.alphabet_, c, rm);
      
      if (++decisions % 10 == 0)
        System.out.println("\t  " + decisions + " clustering decisions.");
      
      best_pos = getBestPositivePair(doc, clustering);
      best_pair = getBestPair(doc, clustering);
    }
    
    // Best pair should be empty.
    if (best_pair != Cluster.EMPTY_PAIR_)
      rm.update(doc, Pair.create(Cluster.EMPTY_PAIR_, best_pair));
    rm.updateAvg();
  }
  
  
  public Pair<Cluster> getBestPositivePair(Document doc, Clustering clustering)
  {
    Discriminant disc = rm_.getDiscriminant();
    
    Pair<Cluster> best_pair = null;
    double best_score = Double.NEGATIVE_INFINITY;

    Vector<Pair<Cluster>> bpairs = new Vector<Pair<Cluster>>();
    bpairs.add(best_pair);

    for (int i = 0; i < clustering.clusters_.size(); i++) {
      Cluster c1 = clustering.clusters_.get(i);
      for (int j = i + 1; j < clustering.clusters_.size(); j++) {
        Cluster c2 = clustering.clusters_.get(j);

        // Temporary distance constraint for efficiency.
        int distance = getElemDistance(c1, c2);
        if (distance > MAX_DISTANCE_)
          continue;
 
        if (c1.getFirstTrueID() == c2.getFirstTrueID()) {
          double score;
          if (clustering.fvs_ != null)
            score = disc.evaluate(clustering.fvs_.get(i, j));
          else {
            // score = rm_.evaluate(doc, Pair.create(c1, c2));
            FeatureVector fv = doc.getFeatureVector(Pair.create(c1, c2), rm_.alphabet_);
            if (fv.features_.size() == 0)
              continue;
            score = rm_.getDiscriminant().evaluate(fv);
          }
          if (score > best_score + EPSILON_) {
            best_pair = new Pair<Cluster>(c1, c2);
            best_score = score;
            bpairs.clear();
            bpairs.add(best_pair);
          }
          else if (score >= best_score - EPSILON_) {
            best_pair = new Pair<Cluster>(c1, c2);
            bpairs.add(best_pair);
          }          
        }
      }
    }
    
    int pos = random_.nextInt(bpairs.size());
    
    return bpairs.get(pos);
  }

  
  public Pair<Cluster> getBestPair(Document doc, Clustering clustering)
  {
    Discriminant disc = rm_.getDiscriminant();
    
    Pair<Cluster> best_pair = Cluster.EMPTY_PAIR_;
    double best_score = rm_.evaluate(doc, best_pair);

    Vector<Pair<Cluster>> bpairs = new Vector<Pair<Cluster>>();
    bpairs.add(best_pair);

    for (int i = 0; i < clustering.clusters_.size(); i++) {
      Cluster c1 = clustering.clusters_.get(i);
      for (int j = i + 1; j < clustering.clusters_.size(); j++) {
        Cluster c2 = clustering.clusters_.get(j);
        
        // Temporary distance constraint for efficiency.
        int distance = getElemDistance(c1, c2);
        if (distance > MAX_DISTANCE_)
          continue;
 
        double score;
        if (clustering.fvs_ != null)
          score = disc.evaluate(clustering.fvs_.get(i, j));
        else {
          // score = rm_.evaluate(doc, Pair.create(c1, c2));
          FeatureVector fv = doc.getFeatureVector(Pair.create(c1, c2), rm_.alphabet_);
          if (fv.features_.size() == 0)
            continue;
          score = rm_.getDiscriminant().evaluate(fv);
        }
        if (score > best_score + EPSILON_) {
          best_pair = new Pair<Cluster>(c1, c2);
          best_score = score;
          bpairs.clear();
          bpairs.add(best_pair);
        }
        else if (score >= best_score - EPSILON_) {
          best_pair = new Pair<Cluster>(c1, c2);
          bpairs.add(best_pair);
        }
      }
    }
    
    int pos = random_.nextInt(bpairs.size());
    
    return bpairs.get(pos);
  }


  public Pair<Cluster> getBestNonEmptyPair(Document doc, Clustering clustering)
  {
    Discriminant disc = rm_.getDiscriminant();
    
    Pair<Cluster> best_pair = null;
    double best_score = Double.NEGATIVE_INFINITY;
    for (int i = 0; i < clustering.clusters_.size(); i++) {
      Cluster c1 = clustering.clusters_.get(i);
      for (int j = i + 1; j < clustering.clusters_.size(); j++) {
        Cluster c2 = clustering.clusters_.get(j);
         
        double score;
        if (clustering.fvs_ != null)
          score = disc.evaluate(clustering.fvs_.get(i, j));
        else
          score = rm_.evaluate(doc, Pair.create(c1, c2));
        if (score > best_score) {
          best_pair = new Pair<Cluster>(c1, c2);
          best_score = score;
        }
      }
    }
    
    return best_pair;
  }

  
  public int getElemDistance(Cluster c1, Cluster c2)
  {
    // Assert elements are sorted inside each cluster.
    int first1 = c1.elements_[0].index_;
    int last1 = c1.elements_[c1.elements_.length - 1].index_;
    
    int first2 = c2.elements_[0].index_;
    int last2 = c2.elements_[c2.elements_.length - 1].index_;
    
    if (first1 < first2) {
      if (last1 < first2)
        return first2 - last1;
      return 0;
    }
    else {
      if (last2 < first1)
        return first1 - last2;
      return 0;
    }
  }
  


}
