package adaptivehac;

import java.util.Iterator;
import java.util.Map;

public class PerceptronRankingModel extends RankingModel
{
  private static final long serialVersionUID = 1L;
  static private double LEARNING_RATE = 1.0; 
  
  transient LinearDiscriminant avg_;
  transient int updates_;
  transient int rounds_;

  public PerceptronRankingModel()
  {
    super();
    disc_ = new LinearDiscriminant();
    avg_ = new LinearDiscriminant();
    updates_ = 0;
    rounds_ = 0;
  }
  
  public void update(Document doc, Pair<? extends Pair<? extends Cluster>> pair)
  {
    Pair<? extends Cluster> positive = pair.getLeft();
    FeatureVector pos_fv = doc.getFeatureVector(positive, alphabet_);
    Pair<? extends Cluster> negative = pair.getRight();
    FeatureVector neg_fv = doc.getFeatureVector(negative, alphabet_);
    
    ((LinearDiscriminant) disc_).add(pos_fv, LEARNING_RATE);
    ((LinearDiscriminant) disc_).sub(neg_fv, LEARNING_RATE);
    
    updates_++;
  }
  
  public void updateAvg()
  {
    avg_.add((LinearDiscriminant) disc_);
    rounds_++;
  }
  
  public void setAverageDiscriminant()
  {
    avg_.scale(1.0 / rounds_);
    disc_ = avg_;
    avg_ = null;
  }
  
  public void printFeatureWeights()
  {
    LinearDiscriminant disc = (LinearDiscriminant) disc_;
    Iterator<Map.Entry<String, Integer>> it = alphabet_.map.entrySet().iterator();
    while (it.hasNext()) {
      Map.Entry<String, Integer> entry = it.next();
      System.out.println(/*entry.getValue() + */" --> " + entry.getKey() + " = " + disc.w_.get(entry.getValue()));
    }
  }

}
