package adaptivehac;

import java.util.Arrays;

public class TriangularMatrix<T>
{
  public int size_;
  public T[] elements_;
  
  @SuppressWarnings("unchecked")
  public TriangularMatrix(int size)
  {
    size_ = size;
    elements_ = (T[]) new Object[size_ * (size_ - 1) / 2];
  }
  
  public T get(int i1, int i2)
  {
    assert(i1 < i2);
    int index = i1 * size_ + i2 - (i1 + 1) * (i1 + 2) / 2;
    
    return elements_[index];
  }

  public void set(int i1, int i2, T t)
  {
    assert(i1 < i2);
    int index = i1 * size_ + i2 - (i1 + 1) * (i1 + 2) / 2;
    
    elements_[index] = t;
  }
  
  public void clear()
  {
    Arrays.fill(elements_, null);
  }

  static public void main(String[] args)
  {
    int size = 10;
    int count = 10;
    
    TriangularMatrix<Integer> tm = new TriangularMatrix<Integer>(size);
    for (int i = 0; i < size; i++) {
      for (int j = i + 1; j < size; j++)
        tm.set(i, j, count++);
    }
    
    for (int i = 0; i < size; i++) {
      for (int j = i + 1; j < size; j++) {
        int c = tm.get(i, j);
        System.out.print(c + " ");
      }
      System.out.println();
    }
   
  }
  
}
